// ChannelDecimateView.cpp : implementation of the CChannelDecimateView class
//

/*
Copyright 2001 Bruce Dawson, Cygnus Software
For more information, visit http://www.cygnus-software.com/papers/
or e-mail: comments@cygnus-software.com

This code may be redistributed as long as this notice is kept intact.
*/

#include "stdafx.h"
#include "ChannelDecimate.h"

#include "ChannelDecimateDoc.h"
#include "ChannelDecimateView.h"

#include "imagelibrary/colourspace.h"
#include <cmath>
#include <vector>
#include <algorithm>
using namespace std;

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CChannelDecimateView

IMPLEMENT_DYNCREATE(CChannelDecimateView, CScrollView)

BEGIN_MESSAGE_MAP(CChannelDecimateView, CScrollView)
	//{{AFX_MSG_MAP(CChannelDecimateView)
	ON_WM_ERASEBKGND()
	ON_COMMAND(ID_VIEW_1XMAGNIFICATION, OnView1xmagnification)
	ON_UPDATE_COMMAND_UI(ID_VIEW_1XMAGNIFICATION, OnUpdateView1xmagnification)
	ON_COMMAND(ID_VIEW_2XMAGNIFICATION, OnView2xmagnification)
	ON_UPDATE_COMMAND_UI(ID_VIEW_2XMAGNIFICATION, OnUpdateView2xmagnification)
	ON_COMMAND(ID_CHANNELS_DECIMATE_INCREASE, OnChannelsDecimateIncrease)
	ON_COMMAND(ID_CHANNELS_DECIMATE_DECREASE, OnChannelsDecimateDecrease)
	ON_COMMAND(ID_VIEW_3XMAGNIFICATION, OnView3xmagnification)
	ON_UPDATE_COMMAND_UI(ID_VIEW_3XMAGNIFICATION, OnUpdateView3xmagnification)
	ON_COMMAND(ID_VIEW_4XMAGNIFICATION, OnView4xmagnification)
	ON_UPDATE_COMMAND_UI(ID_VIEW_4XMAGNIFICATION, OnUpdateView4xmagnification)
	//}}AFX_MSG_MAP
	ON_COMMAND_RANGE(ID_CHANNELS_DECIMATE_1X, ID_CHANNELS_DECIMATE_1024X, OnChannelsDecimatex)
	ON_UPDATE_COMMAND_UI_RANGE(ID_CHANNELS_DECIMATE_1X, ID_CHANNELS_DECIMATE_1024X, OnUpdateChannelsDecimatex)
	ON_COMMAND_RANGE(ID_CHANNELS_Y, ID_CHANNELS_NONE, OnChannelsY)
	ON_UPDATE_COMMAND_UI_RANGE(ID_CHANNELS_Y, ID_CHANNELS_NONE, OnUpdateChannelsY)
	// Standard printing commands
	ON_COMMAND(ID_FILE_PRINT, CScrollView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_DIRECT, CScrollView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, CScrollView::OnFilePrintPreview)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CChannelDecimateView construction/destruction

CChannelDecimateView::CChannelDecimateView()
{
	m_magnification = 1;
	m_decimationAmount = 3;
	m_filteredColours = k_filterNone;
#ifdef	GDCDEMO
	m_maximized = false;
#endif
}

CChannelDecimateView::~CChannelDecimateView()
{
}

BOOL CChannelDecimateView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return base::PreCreateWindow(cs);
}

void CChannelDecimateView::FilterBitmap()
{
	CChannelDecimateDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	const CSBitmap* pSource = &pDoc->m_bitmap;
	if (!pSource->HasBitmap())
	{
		m_filteredBitmap.SetSize(1, 1, 3);
		return;
	}
	CSBitmap* pDest = &m_filteredBitmap;
	// If we already have a destination bitmap, do nothing.
	if (pDest->HasBitmap())
		return;
	const int width = pSource->GetWidth();
	const int height = pSource->GetHeight();
	const int channels = pSource->GetChannels();
	if (m_filteredColours != k_filterNone)
	{
		m_filteredBitmap.SetSize(width, height, 3);
		vector<float>	tempBuffer(width * height * 3);
		const int tempStride = width * 3;
		RGBQUAD palette[256];
		if (channels == 1)
			pSource->GetPalette(0, 256, palette);
		// First we convert the entire image to the desired format.
		for (int y = 0; y < height; ++y)
		{
			const uint8_t* pSourcePixel = pSource->GetLinePtr(y);
			float* pTempPixel = &tempBuffer[tempStride * y];
			// Convert from RGB to the requested colour space.
			for (int x = 0; x < width; ++x)
			{
				TypeRGB<float> rgbPixel;
				if (channels == 1)
				{
					uint8_t index = pSourcePixel[0];
					rgbPixel = TypeRGB<float>(palette[index].rgbRed, palette[index].rgbGreen, palette[index].rgbBlue);
				}
				else
					rgbPixel = TypeRGB<float>(pSourcePixel[RED_BITMAP_OFFSET], pSourcePixel[GREEN_BITMAP_OFFSET], pSourcePixel[BLUE_BITMAP_OFFSET]);
				switch (m_filteredColours)
				{
					case k_filterY:
					case k_filterU:
					case k_filterV:
					case k_filterUV:
					{
						TypeYUV<float> yuvPixel = rgbPixel.ToYUV();
						pTempPixel[0] = yuvPixel.y;
						pTempPixel[1] = yuvPixel.u;
						pTempPixel[2] = yuvPixel.v;
						break;
					}
					case k_filterI:
					case k_filterQ:
					case k_filterIQ:
					{
						TypeYIQ<float> yiqPixel = rgbPixel.ToYIQ();
						pTempPixel[0] = yiqPixel.y;
						pTempPixel[1] = yiqPixel.i;
						pTempPixel[2] = yiqPixel.q;
						break;
					}
					case k_filterR:
					case k_filterG:
					case k_filterB:
						pTempPixel[0] = rgbPixel.r;
						pTempPixel[1] = rgbPixel.g;
						pTempPixel[2] = rgbPixel.b;
						break;
				}
				pSourcePixel += channels;
				pTempPixel += 3;
			}
		}
		// Then we decimate some types.
		int channelMask = 0;
		switch (m_filteredColours)
		{
			case k_filterY:
			case k_filterR:
				channelMask = 1;
				break;
			case k_filterU:
			case k_filterI:
			case k_filterG:
				channelMask = 2;
				break;
			case k_filterQ:
			case k_filterV:
			case k_filterB:
				channelMask = 4;
				break;
			case k_filterUV:
			case k_filterIQ:
				channelMask = 6;
				break;
		}
		for (int channelOffset = 0; channelOffset < 3; ++channelOffset)
		{
			if ((1 << channelOffset) & channelMask)
			{
				for (int y = 0; y < height; /**/)
				{
					// Handle the first line.
					float* const pTempPixel = &tempBuffer[tempStride * (y++) + channelOffset];
					for (int x = 0; x < width; /**/)
					{
						float source = pTempPixel[3 * x++];
						for (int dec = 1; dec < m_decimationAmount && x < width; ++dec)
							pTempPixel[3 * x++] = source;
					}
					// Handle subsequent lines.
					for (int dec = 1; dec < m_decimationAmount && y < height; ++dec)
					{
						float* pTempDestPixel = &tempBuffer[tempStride * (y++) + channelOffset];
						for (int x = 0; x < width; ++x)
							pTempDestPixel[3 * x] = pTempPixel[3 * x];

					}
				}
			}
		}
		// Then we convert back to RGB.
		for (int y = 0; y < height; ++y)
		{
			uint8_t* pDestPixel = pDest->GetLinePtr(y);
			const float* pTempPixel = &tempBuffer[tempStride * y];
			for (int x = 0; x < width; ++x)
			{
				TypeRGB<float> rgbPixel;
				switch (m_filteredColours)
				{
					case k_filterY:
					case k_filterU:
					case k_filterV:
					case k_filterUV:
					{
						TypeYUV<float> yuvPixel(pTempPixel[0], pTempPixel[1], pTempPixel[2]);
						rgbPixel = yuvPixel.ToRGB();
						break;
					}
					case k_filterI:
					case k_filterQ:
					case k_filterIQ:
					{
						TypeYIQ<float> yiqPixel(pTempPixel[0], pTempPixel[1], pTempPixel[2]);
						rgbPixel = yiqPixel.ToRGB();
						break;
					}
					case k_filterR:
					case k_filterG:
					case k_filterB:
						rgbPixel.r = pTempPixel[0];
						rgbPixel.g = pTempPixel[1];
						rgbPixel.b = pTempPixel[2];
						break;
				}
				// We can easily generate RGB values that are outside of the legal range,
				// so we need to check for and clip these.
				if (rgbPixel.r < 0)
					rgbPixel.r = 0;
				if (rgbPixel.r > 255)
					rgbPixel.r = 255;
				if (rgbPixel.g < 0)
					rgbPixel.g = 0;
				if (rgbPixel.g > 255)
					rgbPixel.g = 255;
				if (rgbPixel.b < 0)
					rgbPixel.b = 0;
				if (rgbPixel.b > 255)
					rgbPixel.b = 255;
				// Stuff the values into the bitmap, with proper rounding please.
				pDestPixel[RED_BITMAP_OFFSET] = uint8_t(rgbPixel.r + 0.5);
				pDestPixel[GREEN_BITMAP_OFFSET] = uint8_t(rgbPixel.g + 0.5);
				pDestPixel[BLUE_BITMAP_OFFSET] = uint8_t(rgbPixel.b + 0.5);
				pDestPixel += 3;
				pTempPixel += 3;
			}
		}
	}
	else
	{
		m_filteredBitmap.SetSize(width, height, channels);
		if (channels == 1)
		{
			RGBQUAD palette[256];
			pSource->GetPalette(0, 256, palette);
			pDest->SetPalette(0, 256, palette);
		}
		for (int y = 0; y < height; ++y)
		{
			const uint8_t* pSourceLine = pSource->GetLinePtr(y);
			uint8_t* pDestLine = pDest->GetLinePtr(y);
			memcpy(pDestLine, pSourceLine, channels * width);
		}
	}
#ifdef	GDCDEMO
	// Crude and ugly code to force the window maximized. Hey - it works.
	if (!m_maximized)
	{
		m_maximized = true;
		ShowWindow(SW_SHOWMAXIMIZED);
		for (CWnd* pParent = GetParent(); pParent; pParent = pParent->GetParent())
		{
			pParent->ShowWindow(SW_SHOWMAXIMIZED);
		}
	}
#endif
}

void CChannelDecimateView::OnUpdate(CView* pSender, LPARAM lHint, CObject* pHint) 
{
	CChannelDecimateDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	const CSBitmap* pSource = &pDoc->m_bitmap;
	if (pSource->HasBitmap())
		SetScrollSizes(MM_TEXT, CSize(pSource->GetWidth() * m_magnification, pSource->GetHeight() * m_magnification));
	else
		SetScrollSizes(MM_TEXT, CSize(1, 1));
}

/////////////////////////////////////////////////////////////////////////////
// CChannelDecimateView drawing

void CChannelDecimateView::OnDraw(CDC* pDC)
{
	CChannelDecimateDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	FilterBitmap();
	const CSBitmap* pBitmap = &m_filteredBitmap;

	// Find total size of drawing area. This is the maximum of the client area
	// and the scroll area.
	CRect clientRect;
	GetClientRect(&clientRect);
	CSize totalSize = GetTotalSize();
	if (totalSize.cx > clientRect.right)
		clientRect.right = totalSize.cx;
	if (totalSize.cy > clientRect.bottom)
		clientRect.bottom = totalSize.cy;

	// Place the bitmap in the center of this area.
	const int bitmapDisplayWidth = pBitmap->GetWidth() * m_magnification;
	const int bitmapDisplayHeight = pBitmap->GetHeight() * m_magnification;
	int xOffset = (clientRect.Width() - bitmapDisplayWidth) / 2;
	int yOffset = (clientRect.Height() - bitmapDisplayHeight) / 2;
	pBitmap->StretchDraw(*pDC, xOffset, yOffset, bitmapDisplayWidth, bitmapDisplayHeight,
				0, 0, pBitmap->GetWidth(), pBitmap->GetHeight());

	// Offset by the scroll position - because clip region's aren't affected by mappings?
	CPoint scrollPos = GetDeviceScrollPosition();
	xOffset -= scrollPos.x;
	yOffset -= scrollPos.y;
	// Clear the areas of the window that are not covered by the game bitmap.
	CRgn clientRegion, bitmapRegion;
	clientRegion.CreateRectRgnIndirect(&clientRect);
	bitmapRegion.CreateRectRgn(xOffset, yOffset, xOffset + bitmapDisplayWidth, yOffset + bitmapDisplayHeight);
	clientRegion.CombineRgn(&clientRegion, &bitmapRegion, RGN_DIFF);
	pDC->SelectClipRgn(&clientRegion);
	CBrush	outline(RGB(192, 192, 192));
	pDC->FillRect(&clientRect, &outline);
	pDC->SelectClipRgn(0);
}

BOOL CChannelDecimateView::OnEraseBkgnd(CDC* pDC) 
{
	// FillOutsideRect() doesn't work because it doesn't understand
	// centered bitmaps. I think it might be possible to get it to
	// work - but I choose not to.
//	CBrush br( GetSysColor( COLOR_WINDOW ) );
//	FillOutsideRect( pDC, &br );
	// Tell Windows not to clear the background. I'll do it.
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CChannelDecimateView printing

BOOL CChannelDecimateView::OnPreparePrinting(CPrintInfo* pInfo)
{
	// default preparation
	return DoPreparePrinting(pInfo);
}

void CChannelDecimateView::OnBeginPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add extra initialization before printing
}

void CChannelDecimateView::OnEndPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add cleanup after printing
}

/////////////////////////////////////////////////////////////////////////////
// CChannelDecimateView diagnostics

#ifdef _DEBUG
void CChannelDecimateView::AssertValid() const
{
	base::AssertValid();
}

void CChannelDecimateView::Dump(CDumpContext& dc) const
{
	base::Dump(dc);
}

CChannelDecimateDoc* CChannelDecimateView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CChannelDecimateDoc)));
	return (CChannelDecimateDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CChannelDecimateView message handlers

void CChannelDecimateView::SetMagnification(int mag)
{
	m_magnification = mag;

	CChannelDecimateDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);
	const CSBitmap* pSource = &pDoc->m_bitmap;
	if (pSource->HasBitmap())
		SetScrollSizes(MM_TEXT, CSize(pSource->GetWidth() * m_magnification, pSource->GetHeight() * m_magnification));
	else
		SetScrollSizes(MM_TEXT, CSize(1, 1));
	Invalidate();
}

void CChannelDecimateView::OnView1xmagnification() 
{
	SetMagnification(1);
}

void CChannelDecimateView::OnUpdateView1xmagnification(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(1 == m_magnification);
}

void CChannelDecimateView::OnView2xmagnification() 
{
	SetMagnification(2);
}

void CChannelDecimateView::OnUpdateView2xmagnification(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(2 == m_magnification);
}

void CChannelDecimateView::OnView3xmagnification() 
{
	SetMagnification(3);
}

void CChannelDecimateView::OnUpdateView3xmagnification(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(3 == m_magnification);
}

void CChannelDecimateView::OnView4xmagnification() 
{
	SetMagnification(4);
}

void CChannelDecimateView::OnUpdateView4xmagnification(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(4 == m_magnification);
}

static const char* const s_filterNames[] =
{
	"Y",
	"U",
	"V",
	"UV",
	"I",
	"Q",
	"IQ",
	"R",
	"G",
	"B",
	"None"
};

void CChannelDecimateView::ForceRecalc()
{
	m_filteredBitmap.FreeBitmap();
	Invalidate();
	UpdateWindow();
	char	buffer[100];
	// I display the current decimation amount as a base 2 logarithm so that
	// it won't be so obvious what the current setting is, but I can still 
	// see it on the titlebar.
	double log2 = log(2);
	double logBase2 = log(m_decimationAmount) / log2;
	sprintf(buffer, "%s(%1.1f)", s_filterNames[m_filteredColours], logBase2);
	CWnd* pWindow = AfxGetMainWnd();
	pWindow->SetWindowText(buffer);
	pWindow->UpdateWindow();
}



static const int s_decimationAmount[] =
{
	1, 2, 3, 4, 8, 16, 32, 64, 128, 256, 512, 1024
};

void CChannelDecimateView::OnChannelsDecimatex(int ID) 
{
	int index = ID - ID_CHANNELS_DECIMATE_1X;
	assert(index >= 0 && index < NUMELEMENTS(s_decimationAmount));
	m_decimationAmount = s_decimationAmount[index];
	ForceRecalc();
}

void CChannelDecimateView::OnUpdateChannelsDecimatex(CCmdUI* pCmdUI) 
{
	const int index = pCmdUI->m_nID - ID_CHANNELS_DECIMATE_1X;
	assert(index >= 0 && index < NUMELEMENTS(s_decimationAmount));
	pCmdUI->SetCheck(s_decimationAmount[index] == m_decimationAmount);
}

void CChannelDecimateView::OnChannelsDecimateIncrease() 
{
	const int* pEnd = s_decimationAmount + NUMELEMENTS(s_decimationAmount);
	const int* pCurrent = find(s_decimationAmount, pEnd, m_decimationAmount);
	assert(pCurrent != pEnd);
	if (pCurrent + 1 < pEnd)
	{
		m_decimationAmount = pCurrent[1];
		ForceRecalc();
	}
}

void CChannelDecimateView::OnChannelsDecimateDecrease() 
{
	const int* pEnd = s_decimationAmount + NUMELEMENTS(s_decimationAmount);
	const int* pCurrent = find(s_decimationAmount, pEnd, m_decimationAmount);
	assert(pCurrent != pEnd);
	if (pCurrent > s_decimationAmount)
	{
		m_decimationAmount = pCurrent[-1];
		ForceRecalc();
	}
}



static const EFilteredColours s_filteredColours[] =
{
	k_filterY,
	k_filterU,
	k_filterV,
	k_filterUV,
	k_filterI,
	k_filterQ,
	k_filterIQ,
	k_filterR,
	k_filterG,
	k_filterB,
	k_filterNone
};

void CChannelDecimateView::OnChannelsY(int ID) 
{
	int index = ID - ID_CHANNELS_Y;
	assert(index >= 0 && index < NUMELEMENTS(s_filteredColours));
	m_filteredColours = s_filteredColours[index];
	ForceRecalc();
}

void CChannelDecimateView::OnUpdateChannelsY(CCmdUI* pCmdUI) 
{
	const int index = pCmdUI->m_nID - ID_CHANNELS_Y;
	assert(index >= 0 && index < NUMELEMENTS(s_filteredColours));
	pCmdUI->SetCheck(s_filteredColours[index] == m_filteredColours);
}
