// CompositeViewerView.cpp : implementation of the CCompositeViewerView class
//

/*
Copyright 2001 Bruce Dawson, Cygnus Software
For more information, visit http://www.cygnus-software.com/papers/
or e-mail: comments@cygnus-software.com

This code may be redistributed as long as this notice is kept intact.
*/

#include "stdafx.h"
#include "CompositeViewer.h"

#include "CompositeViewerDoc.h"
#include "CompositeViewerView.h"

#include <cstring>
#include <vector>
#include <cmath>
#include <sstream>
using namespace std;
#include "imagelibrary/colourspace.h"

#include "imagelibrary/csbitmap.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CCompositeViewerView

IMPLEMENT_DYNCREATE(CCompositeViewerView, CView)

BEGIN_MESSAGE_MAP(CCompositeViewerView, CView)
	//{{AFX_MSG_MAP(CCompositeViewerView)
	ON_COMMAND(ID_FILE_SAVE_AS, OnFileSaveAs)
	ON_WM_ERASEBKGND()
	//}}AFX_MSG_MAP
	// Standard printing commands
	ON_COMMAND(ID_FILE_PRINT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_DIRECT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, CView::OnFilePrintPreview)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CCompositeViewerView construction/destruction

CCompositeViewerView::CCompositeViewerView()
{
	// TODO: add construction code here

}

CCompositeViewerView::~CCompositeViewerView()
{
}

BOOL CCompositeViewerView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return CView::PreCreateWindow(cs);
}

/////////////////////////////////////////////////////////////////////////////
// CCompositeViewerView drawing

const int k_textXOffset = 4;
const int k_textYOffset = 2;

void CCompositeViewerView::RenderBitmap(CSBitmap& bitmap)
{
	CDC	*myContext = CDC::FromHandle(bitmap.GetDrawSurface());
	CBrush	block(RGB(255, 255, 255));
	CRect ClientRect = CRect(0, 0, bitmap.GetWidth(), bitmap.GetHeight());
	myContext->FillRect(&ClientRect, &block);
	RenderDC(myContext, ClientRect);
}

void CCompositeViewerView::RenderDC(CDC* pDC, const CRect& ClientRect)
{
	CCompositeViewerDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	DoubleRGB	RGB = pDoc->GetRGB();
	DoubleYIQ	YIQ = RGB.ToYIQ();

	const char* name = 0;
	const int width = ClientRect.Width();
	const int height = ClientRect.Height();

	vector<double> samples(width);
	stringstream overlay;
	/*int oldprecision =*/ overlay.precision(3);
	overlay << "RGB: " << RGB << "\n";
	overlay << "YIQ: " << YIQ << "\n";
	double cAmplitude = sqrt(YIQ.i * YIQ.i + YIQ.q * YIQ.q);
	overlay << "Composite amplitude: " << YIQ.y - cAmplitude << " to " << YIQ.y + cAmplitude << "\n";

	const int numWaveForms = 2;
	const double pi = atan(1) * 4;
	const double xMultiplier = (pi * 2 * numWaveForms) / width;
	bool clip = pDoc->GetClip();
	double minimum = -0.2;	// 0.6;
	double maximum = 1.2;
	for (int x = 0; x < width; ++x)
	{
		double theta = xMultiplier * x;
		samples[x] = YIQ.y + YIQ.i * cos(theta) + YIQ.q * sin(theta);
		if (clip)
		{
			if (samples[x] < minimum)
				samples[x] = minimum;
			if (samples[x] > maximum)
				samples[x] = maximum;
		}
	}

	CPen	BlackPen(PS_SOLID, 4, RGB(0, 0, 0));	//RGB(.65 * 255, .19 * 255, .87 * 255));
	CPen*	oldPen = pDC->SelectObject(&BlackPen);

	const double yMultiplier = height * (pDoc->GetShowDetails() ? 0.50 : 0.65);//0.58);
	const int yOffset = int(height * (pDoc->GetShowDetails() ? 0.60 : 0.50)) + yMultiplier / 2;

	CPen	DottedGraphPen(PS_DOT, 0, RGB(220, 170, 170));
	pDC->SelectObject(&DottedGraphPen);
	for (int i = -2; i <= 12; ++i)
	{
		int y = int(yOffset - (i * 0.1) * yMultiplier);
		pDC->MoveTo(0, y);
		pDC->LineTo(width, y);
	}

	COLORREF GraphColour = RGB(200, 0, 0);
	const int lines[] =
	{
		yOffset,
		//int(yOffset - yMultiplier / 2),
		int(yOffset - yMultiplier),
	};
	CPen	GraphPen(PS_SOLID, 2, GraphColour);
	pDC->SelectObject(&GraphPen);
	for (int i = 0; i < NUMELEMENTS(lines); ++i)
	{
		pDC->MoveTo(0, lines[i]);
		pDC->LineTo(width, lines[i]);
	}


	pDC->SelectObject(&BlackPen);
	pDC->MoveTo(0, int(yOffset - samples[0] * yMultiplier));
	for (int x = 1; x < width; ++x)
	{
		pDC->LineTo(x, int(yOffset - samples[x] * yMultiplier));
	}

	if (pDoc->GetShowDetails())
	{
		string text = overlay.str();
		CRect RectCopy = ClientRect;
		RectCopy.left += k_textXOffset;
		RectCopy.top += k_textYOffset;
		pDC->DrawText(text.c_str(), text.size(), &RectCopy, DT_LEFT);

		CBrush	block(RGB(int(RGB.r * 255 + 0.5), int(RGB.g * 255 + 0.5), int(RGB.b * 255 + 0.5)));
		//CBrush* oldBrush = pDC->SelectObject(&block);
		RectCopy = ClientRect;
		RectCopy.left = RectCopy.right - int(width * 0.3);
		RectCopy.bottom = RectCopy.top + int(height * 0.15);
		pDC->FillRect(&RectCopy, &block);
	}

	//pDC->SelectObject(oldBrush);
	pDC->SelectObject(oldPen);
}

void CCompositeViewerView::OnDraw(CDC* pDC)
{
	CCompositeViewerDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	CRect ClientRect;
	GetClientRect(&ClientRect);
	if (true)	//pDOC->GetDoubleBuffered())
	{
		CSBitmap	TempBitmap;

		TempBitmap.SetSize(ClientRect.Width(), ClientRect.Height(), 3);
		RenderBitmap(TempBitmap);
		TempBitmap.DrawAll(*pDC);
	}
	else
		RenderDC(pDC, ClientRect);
#ifdef	GDCDEMO
	// Crude and ugly code to force the window maximized. Hey - it works.
	static bool m_maximized;
	if (!m_maximized)
	{
		m_maximized = true;
		ShowWindow(SW_SHOWMAXIMIZED);
		for (CWnd* pParent = GetParent(); pParent; pParent = pParent->GetParent())
		{
			pParent->ShowWindow(SW_SHOWMAXIMIZED);
		}
	}
#endif
}

/////////////////////////////////////////////////////////////////////////////
// CCompositeViewerView printing

BOOL CCompositeViewerView::OnPreparePrinting(CPrintInfo* pInfo)
{
	// default preparation
	return DoPreparePrinting(pInfo);
}

void CCompositeViewerView::OnBeginPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add extra initialization before printing
}

void CCompositeViewerView::OnEndPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add cleanup after printing
}

/////////////////////////////////////////////////////////////////////////////
// CCompositeViewerView diagnostics

#ifdef _DEBUG
void CCompositeViewerView::AssertValid() const
{
	CView::AssertValid();
}

void CCompositeViewerView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

CCompositeViewerDoc* CCompositeViewerView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CCompositeViewerDoc)));
	return (CCompositeViewerDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CCompositeViewerView message handlers

void CCompositeViewerView::OnFileSaveAs() 
{
	CSBitmap	TempBitmap;

	CRect ClientRect;
	GetClientRect(&ClientRect);
	TempBitmap.SetSize(ClientRect.Width(), ClientRect.Height(), 3);
	RenderBitmap(TempBitmap);
	//SaveBMPFile("c:\\temp\\composite.bmp", &TempBitmap);
}

BOOL CCompositeViewerView::OnEraseBkgnd(CDC* pDC) 
{
	return TRUE;	// Don't clear the flippin' background!
}
