// SetColour.cpp : implementation file
//

/*
Copyright 2001 Bruce Dawson, Cygnus Software
For more information, visit http://www.cygnus-software.com/papers/
or e-mail: comments@cygnus-software.com

This code may be redistributed as long as this notice is kept intact.
*/

#include "stdafx.h"
#include "CompositeViewer.h"
#include "SetColour.h"

#include <algorithm>
#include <cmath>
using namespace std;

typedef TypeRGB<double> DoubleRGB;
typedef	TypeYUV<double> DoubleYUV;
typedef	TypeYIQ<double>	DoubleYIQ;
typedef TypeHSV<double> DoubleHSV;

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// SetColour dialog

const int k_sliderMin = 0;
const int k_sliderMax = 100;

SetColour::SetColour(DoubleRGB* pColour, CWnd* pView)
	: CDialog(SetColour::IDD, 0), m_pColour(pColour), m_pView(pView)
{
	//{{AFX_DATA_INIT(SetColour)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT
}


void SetColour::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(SetColour)
	DDX_Control(pDX, IDC_VALUE, m_value);
	DDX_Control(pDX, IDC_SATURATION, m_saturation);
	DDX_Control(pDX, IDC_HUE, m_hue);
	DDX_Control(pDX, IDC_RED, m_red);
	DDX_Control(pDX, IDC_GREEN, m_green);
	DDX_Control(pDX, IDC_BLUE, m_blue);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(SetColour, CDialog)
	//{{AFX_MSG_MAP(SetColour)
	ON_WM_HSCROLL()
	ON_BN_CLICKED(IDC_FIXLUMINANCE, OnFixluminance)
	ON_BN_CLICKED(IDC_FIXSATURATION, OnFixsaturation)
	ON_BN_CLICKED(IDC_UNDO, OnUndo)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// SetColour message handlers

BOOL SetColour::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	m_red.SetRange(k_sliderMin, k_sliderMax);
	m_green.SetRange(k_sliderMin, k_sliderMax);
	m_blue.SetRange(k_sliderMin, k_sliderMax);

	m_hue.SetRange(0, 359);
	m_saturation.SetRange(k_sliderMin, k_sliderMax);
	m_value.SetRange(k_sliderMin, k_sliderMax);

	UpdateAll();
	m_backup = m_pColour[0];

	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void SetColour::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
	CDialog::OnHScroll(nSBCode, nPos, pScrollBar);

	static bool s_tracking = false;
	switch (nSBCode)
	{
		case SB_THUMBTRACK:
			if (!s_tracking)
			{
				m_backup = m_pColour[0];
				s_tracking = true;
				OutputDebugString("Backup\n");
			}
			break;
		case SB_THUMBPOSITION:
		case SB_ENDSCROLL:
			OutputDebugString("No backup\n");
			s_tracking = false;
			break;
		default:
			OutputDebugString("Other backup\n");
			m_backup = m_pColour[0];
			break;
	}

	switch (pScrollBar->GetDlgCtrlID())
	{
		case IDC_RED:
		case IDC_GREEN:
		case IDC_BLUE:
			m_pColour->r = double(m_red.GetPos() - k_sliderMin) / (k_sliderMax - k_sliderMin);
			m_pColour->g = double(m_green.GetPos() - k_sliderMin) / (k_sliderMax - k_sliderMin);
			m_pColour->b = double(m_blue.GetPos() - k_sliderMin) / (k_sliderMax - k_sliderMin);
			SetHSVScrollbars();
			break;

		case IDC_HUE:
		case IDC_SATURATION:
		case IDC_VALUE:
		{
			DoubleHSV HSV;
			HSV.h = double(m_hue.GetPos());
			HSV.s = double(m_saturation.GetPos() - k_sliderMin) / (k_sliderMax - k_sliderMin);
			HSV.v = double(m_value.GetPos() - k_sliderMin) / (k_sliderMax - k_sliderMin);
			m_pColour[0] = HSV.ToRGB();
			SetRGBScrollbars();
			break;
		}
	}
	UpdateText();
	m_pView->Invalidate();
}


const double pi = 3.14159265358979323846;
const double cos33 = cos(33 * pi / 180);
const double sin33 = sin(33 * pi / 180);

// Specify the maximum amount you are willing to go
// above 1.0 and below 0.0. This should be set no
// higher than 0.2. 0.1 may produce better results.
// This value assumes RGB values from 0.0 to 1.0.
const double k_maxExcursion = 0.2;

// Returns zero if all is well. If the signal will go too high it returns how
// high it will go. If the signal will go too low it returns how low it
// will go - a negative number.
// It is not possible for the composite signal to simultaneously
// go off the top and bottom of the chart.
// This code is non-optimized for clarity.
double CalcOverheatAmount(double r, double g, double b)
{
	assert(r >= 0.0 && r <= 1.0);
	assert(g >= 0.0 && g <= 1.0);
	assert(b >= 0.0 && b <= 1.0);
	// Convert from RGB to YUV space.
	double y = .299 * r + .587 * g + .114 * b;
	double u = 0.492 * (b - y);
	double v = 0.877 * (r - y);

	// Convert from YUV to YIQ space. This could be combined with
	// the RGB to YIQ conversion for better performance.
	double i = cos33 * v - sin33 * u;
	double q = sin33 * v + cos33 * u;

	// Calculate the amplitude of the chroma signal.
	double c = sqrt(i * i + q * q);
	// See if the composite signal will go too high or too low.
	double maxComposite = y + c;
	double minComposite = y - c;
	if (maxComposite > 1.0 + k_maxExcursion)
		return maxComposite;
	if (minComposite < -k_maxExcursion)
		return minComposite;
	return 0;
}

// This epsilon value is necessary because the standard colour space conversion
// factors are only accurate to three decimal places.
const double k_epsilon = 0.001;

void FixAmplitude(double& r, double& g, double& b)
{
	// maxComposite is slightly misnamed - it may be min or max.
	double maxComposite = CalcOverheatAmount(r, g, b);
	if (maxComposite != 0.0)
	{
		double coolant;
		if (maxComposite > 0)
		{
			// Calculate the ratio between our maximum composite
			// signal level and our maximum allowed level.
			coolant = (1.0 + k_maxExcursion - k_epsilon) / maxComposite;
		}
		else
		{
			// Calculate the ratio between our minimum composite
			// signal level and our minimum allowed level.
			coolant = (k_maxExcursion - k_epsilon) / -maxComposite;
		}
		// Scale R, G and B down. This will move the composite signal
		// proportionally closer to zero.
		r *= coolant;
		g *= coolant;
		b *= coolant;
	}
}

void FixSaturation(double& r, double& g, double& b)
{
	// maxComposite is slightly misnamed - it may be min
	// composite or maxComposite.
	double maxComposite = CalcOverheatAmount(r, g, b);
	if (maxComposite != 0.0)
	{
		// Convert into YIQ space, and convert c, the amplitude
		// of the chroma component.
		DoubleYIQ YIQ = DoubleRGB(r, g, b).ToYIQ();
		double c = sqrt(YIQ.i * YIQ.i + YIQ.q * YIQ.q);
		double coolant;
		// Calculate the ratio between the maximum chroma range allowed
		// and the current chroma range. 
		if (maxComposite > 0)
		{
			// The maximum chroma range is the maximum composite value
			// minus the luminance.
			coolant = (1.0 + k_maxExcursion - k_epsilon - YIQ.y) / c;
		}
		else
		{
			// The maximum chroma range is the luminance minus the
			// minimum composite value.
			coolant = (YIQ.y - -k_maxExcursion - k_epsilon) / c;
		}
		// Scale I and Q down, thus scaling chroma down and reducing the
		// saturation proportionally.
		YIQ.i *= coolant;
		YIQ.q *= coolant;
		DoubleRGB RGB = YIQ.ToRGB();
		r = RGB.r;
		g = RGB.g;
		b = RGB.b;
	}
}

void SetColour::OnFixluminance() 
{
	m_backup = m_pColour[0];
	FixAmplitude(m_pColour->r, m_pColour->g, m_pColour->b);
	UpdateAll();
}

void SetColour::OnFixsaturation() 
{
	m_backup = m_pColour[0];
	FixSaturation(m_pColour->r, m_pColour->g, m_pColour->b);
	UpdateAll();
}

void SetColour::OnUndo() 
{
	swap(m_pColour[0], m_backup);
	UpdateAll();
}

void SetColour::UpdateAll()
{
	UpdateText();
	SetRGBScrollbars();
	SetHSVScrollbars();
	m_pView->Invalidate();
}

void SetColour::SetDlgItemFloat(int id, double value)
{
	char	buffer[100];
	sprintf(buffer, "%1.3f", value);
	SetDlgItemText(id, buffer);
}

void SetColour::UpdateText()
{
	SetDlgItemFloat(IDC_REDTEXT, m_pColour->r);
	SetDlgItemFloat(IDC_GREENTEXT, m_pColour->g);
	SetDlgItemFloat(IDC_BLUETEXT, m_pColour->b);

	DoubleYIQ YIQ = m_pColour->ToYIQ();
	SetDlgItemFloat(IDC_YTEXT, YIQ.y);
	SetDlgItemFloat(IDC_ITEXT, YIQ.i);
	SetDlgItemFloat(IDC_QTEXT, YIQ.q);

	DoubleHSV HSV = m_pColour->ToHSV();
	SetDlgItemFloat(IDC_HUETEXT, HSV.h);
	SetDlgItemFloat(IDC_SATURATIONTEXT, HSV.s);
	SetDlgItemFloat(IDC_VALUETEXT, HSV.v);

	bool hotColour = CalcOverheatAmount(m_pColour->r, m_pColour->g, m_pColour->b) != 0;
	GetDlgItem(IDC_FIXLUMINANCE)->EnableWindow(hotColour);
	GetDlgItem(IDC_FIXSATURATION)->EnableWindow(hotColour);
}

void SetColour::SetRGBScrollbars()
{
	m_red.SetPos(k_sliderMin + int((k_sliderMax - k_sliderMin) * m_pColour->r + 0.5));
	m_green.SetPos(k_sliderMin + int((k_sliderMax - k_sliderMin) * m_pColour->g + 0.5));
	m_blue.SetPos(k_sliderMin + int((k_sliderMax - k_sliderMin) * m_pColour->b + 0.5));
}

void SetColour::SetHSVScrollbars()
{
	DoubleHSV HSV = m_pColour->ToHSV();

	m_hue.SetPos(int(HSV.h));
	m_saturation.SetPos(k_sliderMin + int((k_sliderMax - k_sliderMin) * HSV.s + 0.5));
	m_value.SetPos(k_sliderMin + int((k_sliderMax - k_sliderMin) * HSV.v + 0.5));
}
