// InterlaceSimView.cpp : implementation of the CInterlaceSimView class
//

/*
Copyright 2001 Bruce Dawson, Cygnus Software
For more information, visit http://www.cygnus-software.com/papers/
or e-mail: comments@cygnus-software.com

This code may be redistributed as long as this notice is kept intact.
*/

#include "stdafx.h"
#include "InterlaceSim.h"

#include "InterlaceSimDoc.h"
#include "InterlaceSimView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CInterlaceSimView

IMPLEMENT_DYNCREATE(CInterlaceSimView, CScrollView)

BEGIN_MESSAGE_MAP(CInterlaceSimView, CScrollView)
	//{{AFX_MSG_MAP(CInterlaceSimView)
	ON_WM_ERASEBKGND()
	ON_COMMAND(ID_VIEW_INTERLACE, OnViewInterlace)
	ON_UPDATE_COMMAND_UI(ID_VIEW_INTERLACE, OnUpdateViewInterlace)
	ON_COMMAND(ID_VIEW_NEXTFIELD, OnViewNextfield)
	ON_UPDATE_COMMAND_UI(ID_VIEW_NEXTFIELD, OnUpdateViewNextfield)
	//}}AFX_MSG_MAP
	// Standard printing commands
	ON_COMMAND_RANGE(ID_VIEW_1XMAGNIFICATION, ID_VIEW_16XMAGNIFICATION, OnViewxmagnification)
	ON_UPDATE_COMMAND_UI_RANGE(ID_VIEW_1XMAGNIFICATION, ID_VIEW_16XMAGNIFICATION, OnUpdateViewxmagnification)
	ON_COMMAND(ID_FILE_PRINT, CScrollView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_DIRECT, CScrollView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, CScrollView::OnFilePrintPreview)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CInterlaceSimView construction/destruction

CInterlaceSimView::CInterlaceSimView()
{
	m_magnification = 1;
	m_field = 0;
	m_interlaced = false;
#ifdef	GDCDEMO
	m_maximized = false;
#endif
}

CInterlaceSimView::~CInterlaceSimView()
{
}

BOOL CInterlaceSimView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return CScrollView::PreCreateWindow(cs);
}

void CInterlaceSimView::OnUpdate(CView* pSender, LPARAM lHint, CObject* pHint) 
{
	CInterlaceSimDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	const CSBitmap* pSource = &pDoc->m_bitmap;
	if (pSource->HasBitmap())
		SetScrollSizes(MM_TEXT, CSize(pSource->GetWidth() * m_magnification, pSource->GetHeight() * m_magnification));
	else
		SetScrollSizes(MM_TEXT, CSize(1, 1));
}

/////////////////////////////////////////////////////////////////////////////
// CInterlaceSimView drawing

void CInterlaceSimView::OnDraw(CDC* pDC)
{
	CInterlaceSimDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	CRect clientRect;
	GetClientRect(&clientRect);
	CSize totalSize = GetTotalSize();
	if (totalSize.cx > clientRect.right)
		clientRect.right = totalSize.cx;
	if (totalSize.cy > clientRect.bottom)
		clientRect.bottom = totalSize.cy;

	const CSBitmap* pSource = &pDoc->m_bitmap;
	if (pSource->HasBitmap())
	{
		const int width = pSource->GetWidth();
		const int height = pSource->GetHeight();
		const int channels = pSource->GetChannels();
		CSBitmap	temp;
		temp.SetSize(width, height, channels);
		if (channels == 1)
		{
			RGBQUAD palette[256];
			pSource->GetPalette(0, 256, palette);
			temp.SetPalette(0, 256, palette);
		}
		for (int y = 0; y < height; ++y)
		{
			uint8_t* pDestLine = temp.GetLinePtr(y);
			if (m_interlaced && (y & 1) == m_field)
				memset(pDestLine, 0, width * channels);
			else
			{
				const uint8_t* pSourceLine = pSource->GetLinePtr(y);
				memcpy(pDestLine, pSourceLine, width * channels);
			}
		}

		// Place the bitmap in the center of this area.
		const int bitmapDisplayWidth = pSource->GetWidth() * m_magnification;
		const int bitmapDisplayHeight = pSource->GetHeight() * m_magnification;
		int xOffset = (clientRect.Width() - bitmapDisplayWidth) / 2;
		int yOffset = (clientRect.Height() - bitmapDisplayHeight) / 2;
		temp.StretchDraw(*pDC, xOffset, yOffset, bitmapDisplayWidth, bitmapDisplayHeight,
					0, 0, pSource->GetWidth(), pSource->GetHeight());

		// Offset by the scroll position - because clip region's aren't affected by mappings?
		CPoint scrollPos = GetDeviceScrollPosition();
		xOffset -= scrollPos.x;
		yOffset -= scrollPos.y;

		// Clear the areas of the window that are not covered by the game bitmap.
		CRgn clientRegion, bitmapRegion;
		clientRegion.CreateRectRgnIndirect(&clientRect);
		bitmapRegion.CreateRectRgn(xOffset, yOffset, xOffset + bitmapDisplayWidth, yOffset + bitmapDisplayHeight);
		clientRegion.CombineRgn(&clientRegion, &bitmapRegion, RGN_DIFF);
		pDC->SelectClipRgn(&clientRegion);
	}
	CBrush	outline(RGB(192, 192, 192));
	pDC->FillRect(&clientRect, &outline);
	pDC->SelectClipRgn(0);
#ifdef	GDCDEMO
	// Crude and ugly code to force the window maximized. Hey - it works.
	if (!m_maximized)
	{
		m_maximized = true;
		ShowWindow(SW_SHOWMAXIMIZED);
		for (CWnd* pParent = GetParent(); pParent; pParent = pParent->GetParent())
		{
			pParent->ShowWindow(SW_SHOWMAXIMIZED);
		}
	}
#endif
}

/////////////////////////////////////////////////////////////////////////////
// CInterlaceSimView printing

BOOL CInterlaceSimView::OnPreparePrinting(CPrintInfo* pInfo)
{
	// default preparation
	return DoPreparePrinting(pInfo);
}

void CInterlaceSimView::OnBeginPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add extra initialization before printing
}

void CInterlaceSimView::OnEndPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add cleanup after printing
}

/////////////////////////////////////////////////////////////////////////////
// CInterlaceSimView diagnostics

#ifdef _DEBUG
void CInterlaceSimView::AssertValid() const
{
	CScrollView::AssertValid();
}

void CInterlaceSimView::Dump(CDumpContext& dc) const
{
	CScrollView::Dump(dc);
}

CInterlaceSimDoc* CInterlaceSimView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CInterlaceSimDoc)));
	return (CInterlaceSimDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CInterlaceSimView message handlers

BOOL CInterlaceSimView::OnEraseBkgnd(CDC* pDC) 
{
	return TRUE;
}

void CInterlaceSimView::OnViewInterlace() 
{
	m_interlaced = !m_interlaced;
	Invalidate();
}

void CInterlaceSimView::OnUpdateViewInterlace(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(m_interlaced);
}

void CInterlaceSimView::OnViewNextfield() 
{
	++m_field;
	m_field &= 1;
	Invalidate();
}

void CInterlaceSimView::OnUpdateViewNextfield(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_interlaced);
}

void CInterlaceSimView::OnViewxmagnification(int ID) 
{
	int offset = (ID - ID_VIEW_1XMAGNIFICATION);
	m_magnification = 1 << offset;
	assert(offset >= 0 && offset < 5);

	CInterlaceSimDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);
	const CSBitmap* pSource = &pDoc->m_bitmap;
	if (pSource->HasBitmap())
		SetScrollSizes(MM_TEXT, CSize(pSource->GetWidth() * m_magnification, pSource->GetHeight() * m_magnification));
	else
		SetScrollSizes(MM_TEXT, CSize(1, 1));
	Invalidate();
}

void CInterlaceSimView::OnUpdateViewxmagnification(CCmdUI* pCmdUI) 
{
	int offset = 0;
	int magCopy = m_magnification;
	while (magCopy > 1)
	{
		magCopy >>= 1;
		++offset;
	}
	unsigned int ID = offset + ID_VIEW_1XMAGNIFICATION;
	pCmdUI->SetCheck(ID == pCmdUI->m_nID);
}
