// InterlaceSimDDraw.cpp : Defines the entry point for the application.
//

#include "stdafx.h"

#include <ddraw.h>
#include <commdlg.h>
#include <string>
using namespace std;
#include "imagelibrary/csbitmap.h"

// For DirectDrawCreate
#pragma comment(lib, "ddraw.lib")

//-----------------------------------------------------------------------------
// Local definitions
//-----------------------------------------------------------------------------
const char* programName = "Direct Draw Interlace simulator";

static char s_imageFilename[MAX_PATH] = "interlaceimage.bmp";

//-----------------------------------------------------------------------------
// Global data
//-----------------------------------------------------------------------------

const int k_screenWidth = 800;
const int k_screenHeight = 600;

static int s_flipDelay = 5;
static int s_lineHeight = 8;
static bool s_paused = true;
static bool s_interlace = false;
static int s_phase = 0;
static CSBitmap s_bitmap;

LPDIRECTDRAW                g_pDD = NULL;        // DirectDraw object
LPDIRECTDRAWSURFACE         g_pDDSPrimary = NULL;// DirectDraw primary surface
LPDIRECTDRAWSURFACE         g_pDDSBack = NULL;   // DirectDraw back surface
LPDIRECTDRAWSURFACE         g_pDDSOne = NULL;    // Offscreen surface 1
LPDIRECTDRAWSURFACE         g_pDDSTwo = NULL;    // Offscreen surface 2
LPDIRECTDRAWPALETTE         g_pDDPal = NULL;     // The primary surface palette
bool                        g_bActive = false;   // Is application active?


//-----------------------------------------------------------------------------
// Name: ReleaseAllObjects()
// Desc: Finished with all objects we use; release them
//-----------------------------------------------------------------------------
static void
ReleaseAllObjects(void)
{
    if (g_pDD != NULL)
    {
        if (g_pDDSPrimary != NULL)
        {
            g_pDDSPrimary->Release();
            g_pDDSPrimary = NULL;
        }
        if (g_pDDSOne != NULL)
        {
            g_pDDSOne->Release();
            g_pDDSOne = NULL;
        }
        if (g_pDDSTwo != NULL)
        {
            g_pDDSTwo->Release();
            g_pDDSTwo = NULL;
        }
        if (g_pDDPal != NULL)
        {
            g_pDDPal->Release();
            g_pDDPal = NULL;
        }
        g_pDD->Release();
        g_pDD = NULL;
    }
}




//-----------------------------------------------------------------------------
// Name: InitFail()
// Desc: This function is called if an initialization function fails
//-----------------------------------------------------------------------------
HRESULT
InitFail(HWND hWnd, HRESULT hRet, LPCTSTR szError,...)
{
    char                        szBuff[1024];
    va_list                     vl;

    va_start(vl, szError);
    vsprintf(szBuff, szError, vl);
    ReleaseAllObjects();
    MessageBox(hWnd, szBuff, programName, MB_OK);
    DestroyWindow(hWnd);
    va_end(vl);
    return hRet;
}


void CSBitmapToSurface(IDirectDrawSurface * pdds, CSBitmap& bitmap, int offset, int lineSize)
{
	assert(bitmap.HasBitmap());
	assert(pdds);
	assert(offset == 0 || offset == 1);
    //
    // Make sure this surface is restored.
    //
    pdds->Restore();
    //
    // Get size of the bitmap
    //
    int dx = bitmap.GetWidth();
    int dy = bitmap.GetHeight();
    //
    // Get size of surface.
    //
    DDSURFACEDESC          ddsd = {0};
    ddsd.dwSize = sizeof(ddsd);
    ddsd.dwFlags = DDSD_HEIGHT | DDSD_WIDTH;
    pdds->GetSurfaceDesc(&ddsd);

	HDC hdcImage = bitmap.GetDrawSurface();
	HDC hdc;
    if (pdds->GetDC(&hdc) == DD_OK)
    {
		if (s_interlace)
		{
			for (int y = 0; y < dy && y < int(ddsd.dwHeight); y += lineSize)
			{
				if (((y / lineSize) & 1) == offset)
				{
					StretchBlt(hdc, 0, y, ddsd.dwWidth, lineSize, hdcImage, 0, y, dx, lineSize, SRCCOPY);
				}
				else
				{
					RECT rect = {0, y, ddsd.dwWidth, y + lineSize};
					FillRect(hdc, &rect, HBRUSH(GetStockObject(BLACK_BRUSH)));
				}
			}
		}
		else
		{
			StretchBlt(hdc, 0, 0, ddsd.dwWidth, dy, hdcImage, 0, 0, dx, dy, SRCCOPY);
		}
        pdds->ReleaseDC(hdc);
    }
}


//-----------------------------------------------------------------------------
// Name: InitSurfaces()
// Desc: This function reads the bitmap file FRNTBACK.BMP and stores half of it
//       in offscreen surface 1 and the other half in offscreen surface 2.
// Well, something like that.
//-----------------------------------------------------------------------------
void 
InitSurfaces(void)
{
	CSBitmapToSurface(g_pDDSOne, s_bitmap, 0, s_lineHeight);
	CSBitmapToSurface(g_pDDSTwo, s_bitmap, 1, s_lineHeight);
}




//-----------------------------------------------------------------------------
// Name: RestoreAll()
// Desc: Restore all lost objects
//-----------------------------------------------------------------------------
HRESULT 
RestoreAll(void)
{
    HRESULT                     hRet;

    hRet = g_pDDSPrimary->Restore();
    if (hRet == DD_OK)
    {
        hRet = g_pDDSOne->Restore();
        if (hRet == DD_OK)
        {
            hRet = g_pDDSTwo->Restore();
            if (hRet == DD_OK)
            {
                InitSurfaces();
            }
        }
    }
    return hRet;
}




//-----------------------------------------------------------------------------
// Name: UpdateFrame(int delay)
// Desc: Displays the proper image for the page
//-----------------------------------------------------------------------------
static void
UpdateFrame(int delay)
{
	if (!g_pDD || !g_pDDSBack)
		return;	// Not initialized yet I guess.

    RECT	rcRect;

    rcRect.left = 0;
    rcRect.top = 0;
    rcRect.right = k_screenWidth;
    rcRect.bottom = k_screenHeight;

	for (int i = 0; i < delay; ++i)
	{
		HRESULT hRet = g_pDD->WaitForVerticalBlank( DDWAITVB_BLOCKEND, 0 );
		assert(hRet == DD_OK);
	}

	if (s_phase)
		g_pDDSBack->BltFast(0, 0, g_pDDSOne, &rcRect, FALSE);
	else
		g_pDDSBack->BltFast(0, 0, g_pDDSTwo, &rcRect, FALSE);

	if (!s_paused)
		s_phase = 1 - s_phase;

    // Flip the surfaces
    while (TRUE)
    {
        HRESULT hRet = g_pDDSPrimary->Flip(NULL, 0);
        if (hRet == DD_OK)
            break;
        if (hRet == DDERR_SURFACELOST)
        {
            hRet = RestoreAll();
            if (hRet != DD_OK)
                break;
        }
        if (hRet != DDERR_WASSTILLDRAWING)
            break;
    }
}




//-----------------------------------------------------------------------------
// Name: WindowProc()
// Desc: The Main Window Procedure
//-----------------------------------------------------------------------------
long FAR PASCAL
WindowProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message)
    {
        case WM_ACTIVATE:
            // Pause if minimized
            g_bActive = !((BOOL)HIWORD(wParam));
            return 0L;

        case WM_DESTROY:
            // Clean up and close the app
            ReleaseAllObjects();
            PostQuitMessage(0);
            return 0L;

        case WM_KEYDOWN:
            // Handle any non-accelerated key commands
            switch (wParam)
            {
                case VK_ESCAPE:
                case VK_F12:
                    PostMessage(hWnd, WM_CLOSE, 0, 0);
                    return 0L;
				case VK_UP:
					++s_flipDelay;
					break;
				case VK_DOWN:
					if (s_flipDelay > 1)
						--s_flipDelay;
					break;
				case VK_RIGHT:
					++s_lineHeight;
					InitSurfaces();
					UpdateFrame(0);
					break;
				case VK_LEFT:
					if (s_lineHeight > 1)
					{
						--s_lineHeight;
						InitSurfaces();
						UpdateFrame(0);
					}
					break;
				case 'P':
					s_paused = !s_paused;
					break;
				case 'I':
					s_interlace = !s_interlace;
					InitSurfaces();
					UpdateFrame(0);
					break;
				case 'S':
					if (s_paused)
					{
						s_phase = 1 - s_phase;
						UpdateFrame(0);
					}
					break;
            }
            break;

		case WM_ERASEBKGND:
			return TRUE;	// Stop my window from being erased.

		case WM_SETCURSOR:
            // Turn off the cursor since this is a full-screen app
            SetCursor(NULL);
            return TRUE;
    }
    return DefWindowProc(hWnd, message, wParam, lParam);
}




//-----------------------------------------------------------------------------
// Name: InitApp()
// Desc: Do work required for every instance of the application:
//          Create the window, initialize data
//-----------------------------------------------------------------------------
static HRESULT
InitApp(HINSTANCE hInstance, int nCmdShow)
{
	const char* interlaceClassName = "InterlaceWindowClass";
    // Set up and register window class
	WNDCLASS wc = {0};
    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = WindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInstance;
    wc.hIcon = 0;	//LoadIcon(hInstance, MAKEINTRESOURCE(IDI_MAIN_ICON));
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH )GetStockObject(BLACK_BRUSH);
    wc.lpszMenuName = programName;
    wc.lpszClassName = interlaceClassName;
    verify(RegisterClass(&wc));

    // Create a window
    HWND hWnd = CreateWindowEx(WS_EX_TOPMOST,
                          interlaceClassName,
                          programName,
                          WS_POPUP,
                          0,
                          0,
                          GetSystemMetrics(SM_CXSCREEN),
                          GetSystemMetrics(SM_CYSCREEN),
                          NULL,
                          NULL,
                          hInstance,
                          NULL);
    if (!hWnd)
        return InitFail(0, TRUE, "Couldn't open window");	// Can't return FALSE - that's equal to DD_OK!
    ShowWindow(hWnd, nCmdShow);
    UpdateWindow(hWnd);
    SetFocus(hWnd);

    ///////////////////////////////////////////////////////////////////////////
    // Create the main DirectDraw object
    ///////////////////////////////////////////////////////////////////////////
    HRESULT hRet = DirectDrawCreate(0, &g_pDD, NULL);
    if (hRet != DD_OK)
        return InitFail(hWnd, hRet, "DirectDrawCreateEx FAILED");

    // Get exclusive mode
    hRet = g_pDD->SetCooperativeLevel(hWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN);
    if (hRet != DD_OK)
        return InitFail(hWnd, hRet, "SetCooperativeLevel FAILED");

    // Set the video mode
    hRet = g_pDD->SetDisplayMode(k_screenWidth, k_screenHeight, 24);
	if (hRet != DD_OK)
	    hRet = g_pDD->SetDisplayMode(k_screenWidth, k_screenHeight, 32);
    if (hRet != DD_OK)
        return InitFail(hWnd, hRet, "SetDisplayMode FAILED");

    // Create the primary surface with 1 back buffer
    DDSURFACEDESC ddsd = {0};
    ddsd.dwSize = sizeof(ddsd);
    ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |
                          DDSCAPS_FLIP |
                          DDSCAPS_COMPLEX;
    ddsd.dwBackBufferCount = 1;
    hRet = g_pDD->CreateSurface(&ddsd, &g_pDDSPrimary, NULL);
    if (hRet != DD_OK)
        return InitFail(hWnd, hRet, "CreateSurface FAILED");

    // Get a pointer to the back buffer
    DDSCAPS ddscaps = {0};
    ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
    hRet = g_pDDSPrimary->GetAttachedSurface(&ddscaps, &g_pDDSBack);
    if (hRet != DD_OK)
        return InitFail(hWnd, hRet, "GetAttachedSurface FAILED");

    // Create a offscreen bitmap.
    ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
    ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
    ddsd.dwHeight = k_screenHeight;
    ddsd.dwWidth = k_screenWidth;
    hRet = g_pDD->CreateSurface(&ddsd, &g_pDDSOne, NULL);
    if (hRet != DD_OK)
        return InitFail(hWnd, hRet, "CreateSurface FAILED");

    // Create another offscreen bitmap.
    hRet = g_pDD->CreateSurface(&ddsd, &g_pDDSTwo, NULL);
    if (hRet != DD_OK)
        return InitFail(hWnd, hRet, "CreateSurface FAILED");

    InitSurfaces();

    RECT	rcRect;

    rcRect.left = 0;
    rcRect.top = 0;
    rcRect.right = k_screenWidth;
    rcRect.bottom = k_screenHeight;

	// Flip the surfaces
    while (TRUE)
    {
        hRet = g_pDDSPrimary->Flip(NULL, 0);
        if (hRet == DD_OK)
            break;
        if (hRet == DDERR_SURFACELOST)
        {
            hRet = RestoreAll();
            if (hRet != DD_OK)
                break;
        }
        if (hRet != DDERR_WASSTILLDRAWING)
            break;
    }

    return DD_OK;
}




//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Initialization, message loop
//-----------------------------------------------------------------------------
int PASCAL
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	while (!LoadImageFile(s_imageFilename, &s_bitmap))
	{
		string message = "Couldn't open bitmap '" + string(s_imageFilename) + "'";
		MessageBox(0, message.c_str(), programName, MB_OK);
		OPENFILENAME ofnStruct = {0};
		ofnStruct.lStructSize = sizeof(ofnStruct);
		ofnStruct.lpstrFilter = "*.bmp\0*.bmp\0";
		ofnStruct.lpstrFile = s_imageFilename;
		ofnStruct.nMaxFile = NUMELEMENTS(s_imageFilename);
		ofnStruct.lpstrTitle = programName;
		ofnStruct.Flags = OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST;
		ofnStruct.lpstrDefExt = "bmp";
		if (GetOpenFileName(&ofnStruct) == FALSE)
			return 0;
	}
	if (s_bitmap.GetChannels() < 3)
	{
		// Bump the bitmap up to 24-bit, always.
		assert(s_bitmap.GetChannels() == 1);
		const int width = s_bitmap.GetWidth();
		const int height = s_bitmap.GetHeight();
		CSBitmap tempBitmap24;
		tempBitmap24.SetSize(width, height, 3);
		// The 8-bit bitmap can be copied to the 24-bit bitmap with BitBlt()
		// or with CompositeBitmap() - whichever seems most appropriate.
		BitBlt(tempBitmap24.GetDrawSurface(), 0, 0, width, height, s_bitmap.GetDrawSurface(), 0, 0, SRCCOPY);
		//CompositeBitmap(&tempBitmap24, &tempBitmap, 0, 0);
		s_bitmap.Swap(tempBitmap24);
	}
    if (InitApp(hInstance, nCmdShow) != DD_OK)
        return FALSE;

    while (TRUE)
    {
		MSG	msg;
        if (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
        {
            if (!GetMessage(&msg, NULL, 0, 0))
                return msg.wParam;
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
        else if (g_bActive)
        {
            UpdateFrame(s_flipDelay);
        }
        else
        {
            // Make sure we go to sleep if we have nothing else to do
            WaitMessage();
        }
    }
}
